/*
 * Copyright 2016-2025 NXP
 * All rights reserved.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

/**
 * @file    NXP 4.c
 * @brief   Application entry point.
 */
#include <stdio.h>
#include <math.h>
#include "led.h"
#include "servo.h"
#include "board.h"
#include "peripherals.h"
#include "pin_mux.h"
#include "clock_config.h"
#include "MK66F18.h"
#include "fsl_debug_console.h"
/* TODO: insert other include files here. */

/* TODO: insert other definitions and declarations here. */
volatile static uint32_t millis = 0;

void SysTick_Handler(void){
	millis++;
}

void sleep(uint32_t milli_seconds){
	uint32_t start_millis;
	start_millis = millis;
	while(millis-start_millis < milli_seconds){
	}
}


#define START_DUTY_CYCLE 40
#define ADC_BUFSIZE 1024

volatile bool free_running;
volatile int wait_counter;
volatile bool waiting_for_falling_edge;
volatile int threshold = 1200;

volatile uint8_t current_phase = 0; // Current motor phase (0-2 for 3-phase motor)
volatile uint8_t current_duty_cycle = 80; // Default duty cycle for PWM

uint32_t motor_enable_gpioa_mask[] = {
    BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK | BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK,
    BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK | BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK,
    BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK | BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK
};

uint32_t motor_direction_ftm_channel[] = {
    FTM0_MOTOR_A_FTM0_CH0_CHANNEL,
    FTM0_MOTOR_B_FTM0_CH1_CHANNEL,
    FTM0_MOTOR_C_FTM0_CH2_CHANNEL
};

// Functions
int configure_off(void) {
    uint8_t success = 0;

    // Disable all motor pins
    GPIO_PortClear(BOARD_MOTOR_EN_A_PTA6_GPIO, BOARD_MOTOR_EN_A_PTA6_GPIO_PIN_MASK);
    GPIO_PortClear(BOARD_MOTOR_EN_B_PTA7_GPIO, BOARD_MOTOR_EN_B_PTA7_GPIO_PIN_MASK);
    GPIO_PortClear(BOARD_MOTOR_EN_C_PTA1_GPIO, BOARD_MOTOR_EN_C_PTA1_GPIO_PIN_MASK);

    // Set PWM duty cycle to 0% for all motor pins
    if (FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, FTM0_MOTOR_A_FTM0_CH0_CHANNEL, kFTM_EdgeAlignedPwm, 0) == kStatus_Success &&
        FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, FTM0_MOTOR_B_FTM0_CH1_CHANNEL, kFTM_EdgeAlignedPwm, 0) == kStatus_Success &&
        FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, FTM0_MOTOR_C_FTM0_CH2_CHANNEL, kFTM_EdgeAlignedPwm, 0) == kStatus_Success) {
        success = 1;
    }

    return success;
}

uint8_t reconfigure_for_phase(uint8_t phase, uint8_t duty_cycle) {
    uint8_t success = 0;

    // Turn off all motor pins
    if (configure_off()) {
        // Configure PWM duty cycle for active phase
        if (FTM_UpdatePwmDutycycle(FTM0_PERIPHERAL, motor_direction_ftm_channel[phase], kFTM_EdgeAlignedPwm, duty_cycle) == kStatus_Success) {
            FTM_SetSoftwareTrigger(FTM0_PERIPHERAL, true);

            // Enable H-Bridge for the active phase
            GPIO_PortSet(GPIOA, motor_enable_gpioa_mask[phase]);
            success = 1;
        } else {
            PRINTF("Error: Failed to configure duty cycle for phase %d\n", phase);
        }
    }

    return success;
}

void commute() {
    // Update phase and reconfigure motor
    current_phase = (current_phase - 1 + 3) % 3; // For 3 phases
    reconfigure_for_phase(current_phase, current_duty_cycle);
}

void motor_bemf_irq_handler(void) {
    uint32_t result_values[2] = {0};

    for (int i = 0; i < 2; i++) {
        uint32_t status = ADC16_GetChannelStatusFlags(ADC0_PERIPHERAL, i);
        if (status == kADC16_ChannelConversionDoneFlag) {
            result_values[i] = ADC16_GetChannelConversionValue(ADC0_PERIPHERAL, i);
        }
    }

    if (free_running) {
        if (wait_counter < 2) {
            waiting_for_falling_edge = (result_values[0] >= threshold);
            wait_counter++;
        } else {
            if ((waiting_for_falling_edge && result_values[0] < threshold) || (!waiting_for_falling_edge && result_values[0] > threshold)) {
                commute();
                wait_counter = 0;
                waiting_for_falling_edge = !waiting_for_falling_edge;
            }
        }
    }

#if defined _CORTEX_M && (_CORTEX_M == 4U)
    __DSB();
#endif
}

void bldc_motor_stop() {
    configure_off();
    PRINTF("Motor stopped\n");
}

void bldc_motor_init() {
    current_duty_cycle = START_DUTY_CYCLE;
    reconfigure_for_phase(0, current_duty_cycle); // Hold Phase A and C
	sleep(100); // Wait for alignment
    printf("Motor Initialized\n");
}

void bldc_motor_set_speed(int percentage) {
    if (percentage < 1) {
        bldc_motor_stop();
    } else {
        current_duty_cycle = 40 + (55 * (percentage / 100)); // 10-80% duty cycle range
    }
}

// Define constants
#define IMAGE_SIZE 128
#define THRESHOLD 20
#define SI_CYCLE_LIMIT 200

// Global variables
AT_NONCACHEABLE_SECTION(uint16_t ImageData[IMAGE_SIZE]);
volatile uint16_t currentPixelNumber = 0;
int ImageDataDifference[IMAGE_SIZE];
int CompareValue;
int i;
int BlackLineLeft;
int BlackLineRight;
int RoadMiddle;
int Deviation = 90;

void calculateDeviation(void);
void findBlackLineRight(void);
void findBlackLineLeft(void);

/* ADC1_IRQn interrupt handler */
void ADC1_IRQHANDLER(void) {
    uint32_t result_values[2] = {0};
    for (int i = 0; i < 2; i++) {
        uint32_t status = ADC16_GetChannelStatusFlags(ADC1_PERIPHERAL, i);
        if (status == kADC16_ChannelConversionDoneFlag) {
            result_values[i] = ADC16_GetChannelConversionValue(ADC1_PERIPHERAL, i);
        }
    }
    /* Save the current pixel value to the camera_adc_output array */

    	  ImageData[currentPixelNumber]=result_values[0];
    	  currentPixelNumber++;


    #if defined _CORTEX_M && (_CORTEX_M == 4U)
        __DSB();
    #endif
}

/* FTM1_IRQn interrupt handler */
void FTM1_IRQHANDLER(void) {
  uint32_t intStatus;
  static int16_t parallax1_si_count=0;

  /* Reading all interrupt flags of status register */
  intStatus = FTM_GetStatusFlags(FTM1_PERIPHERAL);
  FTM_ClearStatusFlags(FTM1_PERIPHERAL, intStatus);

  /* We only have a single interrupt source for the FTM1 so we do not need to select (yet) */
  /* This is the counter for the SI cycles, it is counted once per clock cycle, only in the first clock cycle the SI signal is set to high */
  parallax1_si_count++;

  /* After the first cycle, the SI Signal for the Parallax Camera is cleared */
  if(parallax1_si_count==1){
      GPIO_PortClear(BOARD_PARALLAX1_S1_PTA9_GPIO, BOARD_PARALLAX1_S1_PTA9_GPIO_PIN_MASK);
      /* The write pointer into the array is set to zero */
      currentPixelNumber=0;
  }
  /* We start the conversion cycle of the ADC if the si signal is in the 1st to 128th cycle */
  /* Please note that we start triggering the conversion directly after the SI signal has been cleared */
  if((parallax1_si_count >= 1)&&(parallax1_si_count <= 128)){
        ADC16_SetChannelConfig(ADC1_PERIPHERAL, ADC1_CH0_CONTROL_GROUP, &ADC1_channelsConfig[0]);
  }
  /* We reset the SI cycle counter after a certain time period in order to start again another capture of 128 pixels */
  /* The time after the 128th cycle and the restart has to be 20us according to the specification, */
  /* in practice it seems that we can also wait longer and have a longer integration time */
  if(parallax1_si_count>SI_CYCLE_LIMIT){
      parallax1_si_count=0;
      /* As we are resetting the SI count, we also set the SI pin to high */
      GPIO_PortSet(BOARD_PARALLAX1_S1_PTA9_GPIO, BOARD_PARALLAX1_S1_PTA9_GPIO_PIN_MASK);
  }

  /* Add for ARM errata 838869, affects Cortex-M4, Cortex-M4F
     Store immediate overlapping exception return operation might vector to incorrect interrupt. */
  #if defined __CORTEX_M && (__CORTEX_M == 4U)
    __DSB();
  #endif
}


void calculateDeviation() {
    findBlackLineRight();
    findBlackLineLeft();
    RoadMiddle = (BlackLineRight + BlackLineLeft) / 2;
    Deviation = (int)(((float)RoadMiddle / (float)(IMAGE_SIZE - 1)) * 169) + 1;
}

void findBlackLineRight() {
    for (i = 64; i < IMAGE_SIZE - 1; i++) {
        ImageDataDifference[i] = ImageData[i] - ImageData[i + 1];
    }
    CompareValue = THRESHOLD;
    for (i = 64; i < IMAGE_SIZE - 1; i++) {//find biggest difference in order to find the edge of track
        if (ImageDataDifference[i] > CompareValue) {
            CompareValue = ImageDataDifference[i];
            BlackLineRight = i;
        }
    }
}

void findBlackLineLeft() {
    for (i = 65; i > 0; i--) {
        ImageDataDifference[i] = ImageData[i] - ImageData[i - 1];
    }
    CompareValue = THRESHOLD;
    for (i = 65; i > 0; i--) {//find biggest difference in order to find the edge of track
        if (ImageDataDifference[i] > CompareValue) {
            CompareValue = ImageDataDifference[i];
            BlackLineLeft = i;
        }
    }
}

/*
 * @brief   Application entry point.
 */
int main(void) {

    /* Init board hardware. */

	BOARD_InitBootPins();
    BOARD_InitBootClocks();
    BOARD_InitBootPeripherals();
#ifndef BOARD_INIT_DEBUG_CONSOLE_PERIPHERAL
    /* Init FSL debug console. */
    BOARD_InitDebugConsole();
#endif
	SysTick_Config(SystemCoreClock / 1000U);

	PRINTF("Hello World\r\n");

	/*bldc_motor_init();
	commute();
	//bldc_motor_set_speed(60);

	// Simulate motor running by calling commute in a loop
	while (1) {
		commute();
		sleep(15);
		//for (int i = 0; i < 100000; i++) {}
	}*/

	/*//Simulate LED turning on white and then off
	whiteLight();
	sleep(1000);
	turnOff();
	rightLight();
	sleep(1000);
	turnOff();
	leftLight();
	sleep(1000);
	turnOff();
	*/

	/*// Example usage: Turning the car right, left, and then centering
	for(int i = 0; i <= 5; i++) {
		 turnRight();
		 sleep(500);
		 centerServo();
		 sleep(500);
		 turnLeft();
		 sleep(500);
		 centerServo();
		 sleep(500);
		 turnOff();
	 }*/

	//Test for camera
	PRINTF("Camera ADC Test Starting...\r\n");
	whiteLight();

	while (1) {
		// Wait for a full image capture to complete
		if (currentPixelNumber >= IMAGE_SIZE) {
			// Print all pixel values in ImageData
			PRINTF("Image Data ADC Values:\r\n");
			for (int i = 0; i < IMAGE_SIZE; i++) {
				PRINTF("Pixel %d: %d\r\n", i, ImageData[i]);
			}

			// Reset currentPixelNumber to allow a new image capture cycle
			currentPixelNumber = 0;

			sleep(500);
		}
	}

	/*// Initialize the camera array
	for(int i=0; i<128; i++){
		ImageData[i]=1;
	}

	while (1) {
	   sleep(1000);
	   //Print out the values of the array every second
	   for (int i = 0; i < 128; i++) {
		   PRINTF("%3d: %4d  ", i, ImageData[i]);
		   PRINTF("\r\n");
	   }

    }*/


	/*whiteLight();
	//bldc_motor_init();
	while(1){
		//commute();
		calculateDeviation();
		if(RoadMiddle < 64){
			turnRight();
			sleep(15);
		}else if(RoadMiddle > 80){
			turnLeft();
			sleep(15);
		}else{
			centerServo();
			sleep(15);
		};
		printf("%d, %d \r\n", RoadMiddle, Deviation);

	}*/






	return 0;
	}
